package utils

import (
    "os"
    "io"
    "net"
    "time"
    "strconv"
    "math/rand"
    "path/filepath"
    "nexus-server/logger"
    "golang.org/x/crypto/bcrypt"
)

var log = logger.GetInstance()

// Returns the list of subdirectories present in dir.
func GetSubDirs(dir string) ([]string, error) {
    subDirs := []string{}

    currentDir, err := os.Open(dir)
    if err != nil {
        return nil, err
    }

    // Retrieves all files entries in the directory (0 = all files in the directory).
    files, err := currentDir.Readdir(0)
    if err != nil {
        currentDir.Close()
        return nil, err
    }

    currentDir.Close()

    // Loop over file entries
    for _, f := range(files) {
        if f.IsDir() {
            subDirs = append(subDirs, filepath.Join(dir, f.Name()))
        }
    }

    return subDirs, nil
}

// Initializes the random number generator.
func RandInit() {
    rand.Seed(time.Now().UnixNano())
}

// Returns an int in the range [min,max] (both inclusive).
func Rand(min, max int) int {
    return rand.Intn(max-min+1)+min
}

func CopyFiles(source, dest string) error {
    src, err := os.Open(source)
    if err != nil {
        return err
    }
    defer src.Close()

    dst, err := os.Create(dest)
    if err != nil {
        return err
    }
    defer dst.Close()
    _, err = io.Copy(dst, src)

    return err
}

// Returns a hash of the "clear" password passed in argument.
func HashPassword(pwd string) string {
    // Uses bcrypt to hash the password.
    hashedPwd, _ := bcrypt.GenerateFromPassword([]byte(pwd), bcrypt.DefaultCost)
    return string(hashedPwd)
}

// Returns a random MAC address as a string (e.g. "b3:30:49:f8:d0:4b").
func RandMacAddress() (string, error) {
    buf := make([]byte, 6)
    _, err  := rand.Read(buf)
    if err != nil {
        return "", err
    }

    // Sets unicast mode (bit 0 to 0) and locally administered addresses (bit 1 to 1)
    // For more details, see: https://en.wikipedia.org/wiki/MAC_address#Universal_vs._local_(U/L_bit)
    buf[0] &= 0xFE
    buf[0] |= 2

    var mac net.HardwareAddr
    mac = append(mac, buf[0], buf[1], buf[2], buf[3], buf[4], buf[5])
    return mac.String(), nil
}

// Returns true if the specified TCP port is available, false otherwise.
func IsPortAvailable(port int) bool {
    ln, err := net.Listen("tcp", ":" + strconv.Itoa(port))
    if err != nil {
        return false
    }
    ln.Close()
    return true
}
