package vms

import (
    "os"
    "time"
    "errors"
    "path/filepath"
    "encoding/json"
    "nexus-common/template"
    "nexus-server/exec"
    "nexus-server/utils"
    "github.com/google/uuid"
    "github.com/go-playground/validator/v10"
)

type (
    Template struct {
        t template.TemplateSerialized
    }
)

const (
    templateConfFile = "template.json"
    templateDiskFile = "disk.qcow"

    templatePublic  = "public"
    templatePrivate = "private"
)

// Creates a template from a VM's disk and returns a pointer to it.
// Concurrency: safe
func NewTemplateFromVM(name, owner, access string, vm *VM) (*Template, error) {
    vm.mutex.Lock()

    if vm.IsRunning() {
        vm.mutex.Unlock()
        return nil, errors.New("Failed creating template: VM must be stopped")
    }

    // Marks VM as "busy", meaning its disk file is being accessed for a possibly long time.
    if vm.IsDiskBusy() {
        vm.mutex.Unlock()
        return nil, errors.New("Failed creating template: VM disk is in use (busy)")
    }
    vm.DiskBusy = true

    // Clears the VM from being busy.
    defer func(vm *VM) { vm.DiskBusy = false }(vm)

    // Creates the template.
    template, err := newTemplate(name, owner, access)
    if err != nil {
        vm.mutex.Unlock()
        return nil, err
    }

    // Creates the template directory.
    templateDir := template.getTemplateDir()
    if err := os.Mkdir(templateDir, conf.MkDirPerm); err != nil {
        msg := "Failed creating template dir: " + err.Error()
        log.Error(msg)
        vm.mutex.Unlock()
        return nil, errors.New(msg)
    }

    vm.mutex.Unlock()

    // 1) Copies VM's overlay file into a new one (in the VM's directory):
    //    cp disk.qcow disk.tpl.qcow
    vmDiskFile := filepath.Join(vm.dir, vmDiskFile)
    tplDiskFile := filepath.Join(vm.dir, "/disk.tpl.qcow")
    if err := utils.CopyFiles(vmDiskFile, tplDiskFile); err != nil {
        template.delete()
        msg := "Failed copying VM overlay disk: " + err.Error()
        log.Error(msg)
        return nil, errors.New(msg)
    }

    // 2) Rebases the new template overlay in order to become a standalone disk file:
    //    qemu-img rebase -b "" disk.tpl.qcow
    if err := exec.QemuImgRebase(tplDiskFile); err != nil {
        template.delete()
        os.Remove(tplDiskFile)
        return nil, err
    }

    // 3) Moves the template disk file into the template directory:
    //    mv disk.tpl.qcow new_template_dir
    destFile := filepath.Join(templateDir, templateDiskFile)
    if err := os.Rename(tplDiskFile, destFile); err != nil {
        template.delete()
        os.Remove(tplDiskFile)
        msg := "Failed moving template disk image to final location: " + err.Error()
        log.Error(msg)
        return nil, errors.New(msg)
    }

    return template, nil
}

// Creates a template from a qcow file and returns a pointer to it.
// Concurrency: safe
func NewTemplateFromQCOW(name, owner, access, qcowFile string) (*Template, error) {
    // Creates the template.
    template, err := newTemplate(name, owner, access)
    if err != nil {
        return nil, err
    }

    // Creates the template directory.
    templateDir := template.getTemplateDir()
    if err := os.Mkdir(templateDir, conf.MkDirPerm); err != nil {
        msg := "Failed creating template dir: " + err.Error()
        log.Error(msg)
        return nil, errors.New(msg)
    }

    // Moves the qcow file from its location to the template directory.
    if err := os.Rename(qcowFile, filepath.Join(templateDir, templateDiskFile)); err != nil {
        msg := "Failed moving uploaded qcow to template dir: " + err.Error()
        log.Error(msg)
        return nil, errors.New(msg)
    }

    return template, nil
}

// Checks whether the specified access is valid.
// Returns an error if it's not.
// Concurrency: safe
func ValidateTemplateAccess(access string) error {
    if access != templatePrivate && access != templatePublic {
        return errors.New("Wrong template access type")
    }
    return nil
}

// Returns true if the template is public.
// Concurrency: safe
func (template *Template)IsPublic() bool {
    return template.t.Access == templatePublic
}

// Returns the absolute path to the template's disk image.
// Concurrency: safe
func (template *Template)GetTemplateDiskPath() string {
    return filepath.Join(template.getTemplateDir(), templateDiskFile)
}

// Returns the template's owner.
// Concurrency: safe
func (template *Template)GetOwner() string {
    return template.t.Owner
}

// Returns a serialized version of the template.
// Concurrency: safe
func (template *Template)SerializeToNetwork() template.TemplateSerialized {
    return template.t
}

// Creates a template given the specified name, owner and access and returns a pointer to it.
// Concurrency: safe
func newTemplate(name, owner, access string) (*Template, error) {
    id, err := uuid.NewRandom()
    if err != nil {
        log.Error("Failed creating template: " + err.Error())
        return nil, err
    }

    template := &Template { template.TemplateSerialized{id, name, owner, access, time.Now()} }

    if err := template.validate(); err != nil {
        return nil, errors.New("Failed validating template: " + err.Error())
    }

    return template, nil
}

// Creates a template from a template configuration file (json) and returns a pointer to it.
// Concurrency: safe
func newTemplateFromFile(file string) (*Template, error) {
    filein, err := os.OpenFile(file, os.O_RDONLY, 0)
    if err != nil {
        return nil, errors.New("Failed reading template config file: " + err.Error())
    }
    defer filein.Close()

    // Decodes the json file into users, then checks:
    // - all fields are present:
    // - extra fields are not allowed
    decoder := json.NewDecoder(filein)
    decoder.DisallowUnknownFields()
    template := &Template{}
    if err = decoder.Decode(&template.t); err != nil {
        return nil, errors.New("Failed decoding template config file: " + err.Error())
    }

    if err = template.validate(); err != nil {
        return nil, errors.New("Failed validating template config file: " + err.Error())
    }

    return template, nil
}

// Writes a template's configuration file (json).
// Concurrency: unsafe
func (template *Template)writeConfig() error {
    templateDir := template.getTemplateDir()
    templateConfigFile := filepath.Join(templateDir, templateConfFile)

    file, err := os.Create(templateConfigFile)
    if err != nil {
        template.delete()
        msg := "Failed writing template config file: " + err.Error()
        log.Error(msg)
        return errors.New(msg)
    }
    defer file.Close()
    encoder := json.NewEncoder(file)
    encoder.SetIndent("", "    ")
    if err = encoder.Encode(template.t); err != nil {
        template.delete()
        msg := "Failed encoding template config file: " + err.Error()
        log.Error(msg)
        return errors.New(msg)
    }
    return nil
}

// Deletes a template's directory along its content.
// Concurrency: unsafe
func (template *Template)delete() error {
    templateDir := template.getTemplateDir()
    if err := os.RemoveAll(templateDir); err != nil {
        msg := "Failed deleting template files: " + err.Error()
        log.Error(msg)
        return errors.New(msg)
    }
    return nil
}

// Returns a template's directory.
// Concurrency: safe
func (template *Template)getTemplateDir() string {
    return filepath.Join(templates.dir, template.t.ID.String())
}

// Checks that the template's fields are valid.
// Returns an error if they are not.
// Concurrency: safe
func (template *Template)validate() error {
    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(template.t); err != nil {
        return err
    }
    return ValidateTemplateAccess(template.t.Access)
}
