package vms

import (
    "os"
    "sync"
    "time"
    "path"
    "errors"
    "strings"
    "net/mail"
    "path/filepath"
    "encoding/json"
    vmc "nexus-common/vm"
    "nexus-common/caps"
    "nexus-server/exec"
    "github.com/google/uuid"
    "github.com/go-playground/validator/v10"
)

type (
    // Internal VM structure.
    VM struct {
        v vmc.VMDiskSerialized

        // None of the fields below are serialized to disk.
        dir string         // VM directory
        qgaSock string     // QEMU Guest Agent (QGA) UNIX socket
        Run runStates
        DiskBusy bool      // When true, the disk image MUST NOT be modified (e.g. write access)
        mutex *sync.Mutex
    }

    runStates struct {
        State vmc.VMState
        Pid int           // QEMU process ID
        SpicePort int     // Port for Spice access
        SpicePwd string   // Password for Spice access
        AttachPwd string  // Password required to attach to this VM
        StartTime time.Time
    }

    endOfExecCallback func(vm *VM)
)

const (
    vmDiskFile     = "disk.qcow"
    vmConfFile     = "vm.json"
    vmSecretFile   = "secret"
    vmQGASockFile  = "qga.sock"
)

// Creates a VM and returns a pointer to it.
// Concurrency: safe
func NewVM(creatorEmail string, name string, cpus, ram int, nic vmc.NicType, usbDevs []string, templateID uuid.UUID, owner string) (*VM, error) {
    vmID, err := uuid.NewRandom()

    if err != nil {
        log.Error("Failed creating VM: "+err.Error())
        return nil, err
    }
    vm := newEmptyVM()
    vm.v.ID = vmID
    vm.v.Owner = owner
    vm.v.Name = name
    vm.v.Cpus = cpus
    vm.v.Ram = ram
    vm.v.Nic = nic
    vm.v.UsbDevs = usbDevs
    vm.v.TemplateID = templateID
    id := vmID.String()
    vm.dir = filepath.Join(vms.dir, id[0:3], id[3:6], id)
    vm.qgaSock = filepath.Join(vm.dir, vmQGASockFile)
    vm.v.Access = make(map[string]caps.Capabilities)

    if err = vm.validate(); err != nil {
        return nil, errors.New("Failed validating VM: "+err.Error())
    }

    return vm, nil
}

// Returns a serialized version of the VM with fields that must be stored on disk.
// Concurrency: safe
func (vm *VM)SerializeForDisk() vmc.VMDiskSerialized {
    return vmc.VMDiskSerialized {
        ID: vm.v.ID,
        Owner: vm.v.Owner,
        Name: vm.v.Name,
        Cpus: vm.v.Cpus,
        Ram: vm.v.Ram,
        Nic: vm.v.Nic,
        UsbDevs: vm.v.UsbDevs,
        TemplateID: vm.v.TemplateID,
        Access: vm.v.Access,
        CreationTime: vm.v.CreationTime,
    }
}

// Returns a serialized version of the VM with most fields except credentials.
// Concurrency: safe
func (vm *VM)Serialize() vmc.VMNetworkSerialized {
    return vmc.VMNetworkSerialized {
        ID: vm.v.ID,
        Owner: vm.v.Owner,
        Name: vm.v.Name,
        Cpus: vm.v.Cpus,
        Ram: vm.v.Ram,
        Nic: vm.v.Nic,
        UsbDevs: vm.v.UsbDevs,
        TemplateID: vm.v.TemplateID,
        Access: vm.v.Access,
        CreationTime: vm.v.CreationTime,
        State: vm.Run.State,
        StartTime: vm.Run.StartTime,
        DiskBusy: vm.DiskBusy,
    }
}

// Returns a serialized version of the VM with basic fields and credentials.
// Concurrency: safe
func (vm *VM)SerializeAttachCredentials() vmc.VMAttachCredentialsSerialized {
    return vmc.VMAttachCredentialsSerialized {
        ID: vm.v.ID,
        Name: vm.v.Name,
        Pwd: vm.Run.AttachPwd,
    }
}

func (vm *VM)SerializeSpiceCredentials() vmc.VMSpiceCredentialsSerialized {
    return vmc.VMSpiceCredentialsSerialized {
        Name: vm.v.Name,
        SpicePort: vm.Run.SpicePort,
        SpicePwd: vm.Run.SpicePwd,
    }
}

// Returns a VM's ID.
// Concurrency: safe
func (vm *VM)GetID() uuid.UUID {
    return vm.v.ID
}

// Returns a VM's access.
// Concurrency: safe
func (vm *VM)GetAccess() map[string]caps.Capabilities {
    return vm.v.Access
}

// Returns true if the specified email is the VM's owner.
// Concurrency: safe
func (vm *VM)IsOwner(email string) bool {
    return email == vm.v.Owner
}

// Returns true if the VM is running.
// Concurrency: safe
func (vm *VM)IsRunning() bool {
    return vm.Run.State == vmc.StateRunning
}

// Returns true if the VM's disk is being modified (e.g. written).
// Concurrency: safe
func (vm *VM)IsDiskBusy() bool {
    return vm.DiskBusy
}

// Returns the absolute path to the VM's disk image.
// Concurrency: safe
func (vm *VM)getDiskPath() string {
    return filepath.Join(vm.dir, vmDiskFile)
}

// Creates an empty VM and returns a pointer to it.
// Concurrency: safe
func newEmptyVM() *VM {
    return &VM {
        v: vmc.VMDiskSerialized {
            ID: uuid.Nil,
            Owner: "",
            Name: "",
            Cpus: 0,
            Ram: 0,
            Nic: "",
            UsbDevs: []string{},
            TemplateID: uuid.Nil,
            Access: make(map[string]caps.Capabilities),
            CreationTime: time.Now(),
        },
        dir: "",
        qgaSock: "",
        Run: runStates { State: vmc.StateStopped, Pid: 0, SpicePort: 0, SpicePwd: "" },
        DiskBusy: false,
        mutex: new(sync.Mutex),
    }
}

// Creates a VM from a configuration file (json) and returns a pointer to it.
// Concurrency: safe
func newVMFromFile(vmFile string) (*VM, error) {
    filein, err := os.OpenFile(vmFile, os.O_RDONLY, 0)
    if err != nil {
        return nil, errors.New("Failed reading VM config file: "+err.Error())
    }
    defer filein.Close()

    // Decodes and validates the json file
    decoder := json.NewDecoder(filein)
    decoder.DisallowUnknownFields()
    vm := newEmptyVM()
    if err = decoder.Decode(&vm.v); err != nil {
        return nil, errors.New("Failed decoding VM config file: "+err.Error())
    }

    // Backward compatibility code to work with VMs that are missing the CreationTime field.
    // TODO: to be removed once all VMs store their creation time.
    if vm.v.CreationTime.IsZero() {
        vm.v.CreationTime = time.Now()
    }

    if err = vm.validate(); err != nil {
        return nil, errors.New("Failed validating VM: "+err.Error())
    }

    vm.dir = path.Dir(vmFile)
    vm.qgaSock = filepath.Join(vm.dir, vmQGASockFile)
    return vm, nil
}

// Checks that the VM structure's fields are valid.
// Returns an error if they are not.
// Concurrency: safe
func (vm *VM)validate() error {
    // Checks the capabilities are valid
    for email, accessCaps := range vm.v.Access {
        _, err := mail.ParseAddress(email)
        if err != nil {
            return errors.New("Invalid email")
        }
        if err := caps.ValidateVMAccessCaps(accessCaps); err != nil {
            return err
        }
    }

    if (vm.v.Nic != vmc.NicNone && vm.v.Nic != vmc.NicUser) {
        return errors.New("Invalid nic value: "+string(vm.v.Nic))
    }

    if err := validateUsbDevs(vm.v.UsbDevs); err != nil {
        return err
    }

    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(vm.v); err != nil {
        return err
    }

    // Checks that the template referenced by the VM actually exists.
    _, err := GetTemplatesInstance().GetTemplate(vm.v.TemplateID)
    if err != nil {
        return err
    }

    return nil
}

// Creates the VM's directory and writes its files:
// 1) Creates the 3-directory structure based on the VM's ID (UUID).
// 2) Writes the VM configuration file (json).
// 3) Creates the VM disk image as an overlay on top of the template disk image.
// Concurrency: unsafe
func (vm *VM)writeFiles() error {
    // Checks the template referenced by the VM exists.
    template, err := GetTemplatesInstance().GetTemplate(vm.v.TemplateID)
    if err != nil {
        return err
    }

    // Creates the 3-directory structure.
    if err = os.MkdirAll(vm.dir, conf.MkDirPerm); err != nil {
        msg := "Failed creating VM dirs: "+err.Error()
        log.Error(msg)
        return errors.New(msg)
    }

    // Writes vmConfFile.
    if err = vm.writeConfig(); err != nil {
        return err
    }

    // Creates vmDiskFile as an overlay on top of the template disk.
    // REMARK: template and output file must be both specified as absolute paths.
    templateDiskFile := template.GetTemplateDiskPath()

    // The "../../../../" is because the template qcow file must be referenced relatively to the VM qcow file
    // which is deeply burried in the multi-level VM directory structure.
    if err := exec.QemuImgCreate(filepath.Join("../../../../", templateDiskFile), vm.getDiskPath()); err != nil {
        vm.delete()
        return err
    }

    return nil
}

// Writes the VM config file (json).
// REMARK: does not check whether the template is valid or not.
// Concurrency: unsafe
func (vm *VM)writeConfig() error {
    vmFile := filepath.Join(vm.dir, vmConfFile)
     file, err := os.Create(vmFile)
     if err != nil {
         msg := "Failed writing VM config file: "+err.Error()
         log.Error(msg)
        return errors.New(msg)
    }
    defer file.Close()
    encoder := json.NewEncoder(file)
    encoder.SetIndent("", "    ")

    if err = encoder.Encode(vm.SerializeForDisk()); err != nil {
        msg := "Failed encoding VM config file: "+err.Error()
        log.Error(msg)
        return errors.New(msg)
    }

    return nil
}

// Deletes all files associated to a VM.
// Concurrency: unsafe
func (vm *VM)delete() error {
    // Deletes the VM's directory and its content.
    if err := os.RemoveAll(vm.dir); err != nil {
        log.Error("Failed deleting VM files: "+err.Error())
        return errors.New("Failed deleting VM files: "+err.Error())
    }

    // Deletes parents directories if they are empty.
    // Directories are only deleted if empty (otherwise an error is triggered which we ignore).
    parentDir := path.Dir(vm.dir)
    parentParentDir := path.Dir(parentDir)
    os.Remove(parentDir)
    os.Remove(parentParentDir)
    return nil
}

// Validates and convert a string of USB devices of the form "1fc9:001d,067b:2303"
// into a slice of string where each element is a string of the form "1fc9:001d".
// Concurrency: safe
func validateUsbDevs(devs []string) error {
    for _, dev := range devs {
        ids := strings.Split(dev, ":")  // Extracts vendorID (vid) and productID (pid)
        if len(ids) != 2 {  // expect exactly two values: vid and pid
            return errors.New("Invalid USB device syntax")
        }
        vid, pid := ids[0], ids[1]
        if !isUsbId(vid) || !isUsbId(pid) {
            return errors.New("Invalid USB vendor/product ID")
        }
    }
    return nil
}

// A USB ID is either a vendor ID or a product ID.
// It's a string containing exactly 4 hexadecimal digits.
// Concurrency: safe
func isUsbId(s string) (bool) {
    if len(s) != 4 {
        return false
    }
    for _, r := range s {
        if !(r >= '0' && r <= '9' || r >= 'a' && r <= 'f' || r >= 'A' && r <= 'F') {
            return false
        }
    }
    return true
}
